# --*-PERL-*--
#####################################################################
#
# Copyright 1996-98, CyberCash Inc. All rights reserved.
# Written by Judy Grass
#
# These routines are intended to be used exclusively with the 
# CyberCash secure payment system.
#
# RCSID: $Id: CCMerchantCustom.pm,v 1.19 1998/12/11 18:02:40 grass Exp $
#
#   This module conatins some routines used by MCK payment scripts
#   that *may* be customized by merchants for various purposes,
#   but that do not *REQUIRE* any customization.
#
#   These will operate fairly well even if you never look at them.
#
# Do not redistribute without written permission from CyberCash Inc.
######################################################################

package CCMerchantCustom;

require 5.003;

# The following 4 lines are here only because the current directory for
# Microsoft IIS when running script on browser is always the root
# directory of the web server.
BEGIN {
    push(@INC, '/home/httpd/test-store/cgi-bin/mck-cgi');
    push(@INC, '/home/httpd/test-store/cgi-bin/mck-cgi/');
}

require Exporter;
@ISA = (Exporter);
@EXPORT = qw();
@EXPORT_OK = qw( GenerateOrderId
	ValidatePrice SendPayload
	ReadPaymentInfo LoadPaymentInfo);

use CCMerchantTest qw(GetOrderInfo &TemplateFailure);

use CCMckLib3_2 qw(%Config $defaultCurrency Trim ValidateOrderId
	PrintTemplate CCError CCDebug CCDebug2);

use CCMckErrno3_2 qw(MCKGetErrorMessage $E_NoErr
		$E_Bad_Orderid
		$E_No_OrderLog
		$E_Order_NotFound
		$E_No_Amount
		$E_Bad_Amount
		$E_Bad_Currency
		$E_Reject_Currency
		$E_Bad_Payload
		$E_No_Payload
	);

#################################
#
# GenerateOrderId
#
# This subroutine generates an order ID, which
# must be unique. This example bases the order number
# on GMT time and the process ID number ($$)
#
# This is provided for your convenience... 
# You may use it as is or you may customize this 
#
#################################

sub GenerateOrderId {
    my($orderID);
    my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime(time());

    # You can use any scheme to generate order ID you want, 
    # so long as:
    #   1) the order ID is unique,
    #   2) it is 32 characters or less, and
    #   3) the following characters are not used 
    #         : < > = + @ " % = &
    # You could read order IDs from a file or a database, or generate
    # one on the fly as we do here.
    #
    # We'll make an order ID based on date, time, and process number
    #
    # $mon is the month index where Jan=0 and Dec=11, so we use
    # $mon+1 to get the more familiar Jan=1 and Dec=12
    #

	# $year is years since 1900:  we want only two digits
    $orderID = sprintf("%02d%02d%02d%02d%02d%05d",
		       $year % 100, $mon+1,$mday,$hour,$min,$$);

    # The following characters are illegal in an order ID:
    #    : < > = + @ " % = &
    #
    # If you want, you could use a line similar to the following
    # to remove these illegal characters:
    #    $orderID =~ tr/:<>=+\@\"\%\&//d;
    #
    # Or use something like the following line to only allow 
    # alphanumeric and dash, converting other characters to underscore:
    #    $orderID =~ tr/A-Za-z0-9\-/_/c;
    #
    # Our test order ID only contains digits, so we don't have
    # to strip any characters here. You might have to if you
    # use a different scheme.
    #
    return $orderID;
}

##################################
# LoadPaymentInfo
#     LoadPaymentInfo($orderId, \%query, \%paymentNVList);
#
# Inputs:
#     $orderId  -- order identifier... key to recovering payment info
#     $queryP  -- pointer to the CGI input parameters
#     $paymentNVListP  -- pointer to the array of payment
#                          request arguments that we are building
#     $problemTemplate -- template to use for error message if this fails
#
#  Returns %paymentNVList
#          Notifies, logs then exits if there is a problem
#  
#  This collects order related payment info and loads it into
#   the vector of payment arguments (paymentNVList)...
#   This sets a number of required and optional "mo." fields in the
#   payment request message.
#
# Our sample implementation gets the payment data from a flat file log
# It uses an orderId for the key.  This orderId is passed to the CGI as a query
#  parameter 
#
# Your implementation will probably store the payment data some other way.
###########################

sub LoadPaymentInfo
{
	my ($orderId, $queryP, $paymentNVListP, $problemTemplate) = @_;
	my (%query, %paymentNVList);

	my($templatePrms, $errmsg, $errcode);
	my($found, $oid, $timestamp, $price, $orderValidTill, $payload,
        $customerId, $product);

	%query = %$queryP;
	%paymentNVList = %$paymentNVListP;

	#Test this to make sure it's a valid one...

	if (!&ValidateOrderId($orderId)) {
		$errmsg = &MCKGetErrorMessage($E_Bad_Orderid);

		$templatePrms{'MESSAGE'} = $errmsg;
	
		&CCError("$0: $errmsg: $orderId\n");
		$errcode = &PrintTemplate($problemTemplate, %templatePrms);
		if ($errcode != $E_NoErr) {
			&TemplateFailure("$errmsg: $orderId");
		}
		exit 1;  # invalid OID
	}

	$paymentNVList{'mo.order-id'} = $orderId;

	#Now read anything you need from the form into 
	#the appropriate message arguments:

	#Get Order Info from the order log
	#  This uses the test version of an order log
	#  The test version will not scale to large stores.
	#  You will need to replace this function with one that makes 
	#  sense for *your* storefront
	
	($found, $oid, $timestamp, $price, $orderValidTill, $payload,
	        $customerId, $product) = &GetOrderInfo ($orderId);

	if ($found <= 0) {
		# this oid wasn't in the log... you can't continue...
		# Either the input form was tampered with, or you have a fundamental
		# malfunction
	
		if ($found < 0) {  #DB failure....
			$errmsg = &MCKGetErrorMessage($E_No_OrderLog);
			$templatePrms{'MESSAGE'} = $errmsg;
		
			&CCError("$0: $errmsg\n");  #may be fraud...
			$errcode = &PrintTemplate($problemTemplate, %templatePrms);
			if ($errcode != $E_NoErr) {
				&TemplateFailure("$errmsg");
			}
			exit 1;
		}
		else { # -1 indicates order not found
			$errmsg = &MCKGetErrorMessage($E_Order_NotFound) . " ($orderId)";
			$templatePrms{'MESSAGE'} = $errmsg;
		
			&CCError("$0: $errmsg\n");  #may be fraud...
			$errcode = &PrintTemplate($problemTemplate, %templatePrms);
			if ($errcode != $E_NoErr) {
				&TemplateFailure("$errmsg");
			}
			exit 1;
		}
	}

	#At this point we **SHOULD** have a price, etc...
	# Validate the price for syntax and currency choice
	# This also will default the currency if needed
	# and return the currency and amount as separate entities to allow
	# price manipulations...
	#  We use '.' as a decimal point
	#  We expect numbers of the form:
	#     digits . digits
	#     digits
	#             .digits
	#
	#  We do not support thousands separators or localized decimal
	#  point representations

	($errcode, $price, $currency, $amount) = &ValidatePrice($price);
	if ($errcode != $E_NoErr) {
		# bad Price
			$errmsg = &MCKGetErrorMessage($errcode) . "
			($price)";

		$templatePrms{"MESSAGE"} = $errmsg; $errcode =
		&PrintTemplate($problemTemplate, %templatePrms);
			if ($errcode != $E_NoErr) {
				&TemplateFailure("$errmsg"); }

		&CCError("$0: $errmsg\n"); 
		exit 1;
	}

	# If you need to manipulate the price, try this.
	# $currency is a currency string (i.e. "usd")
	# $amount is a NUMERICAL amount value that you can
	# do arithmetic with....
	# 
	# if you want to apply a discount, or adjust the price in
	# some other way, this is the place to do it
	# 
	# $amount is the variable that you want to adjust
	# For example:
	#    $amount = $amount - 1.00;   # take 1.00 off
	#    $amount = $amount * 0.90;   # 10% discount
	#
	# Now, reformat the price correctly
	#
	#  $amount = sprintf("%.2f",$amount);
	#  $price = sprintf("%3s %.2f",$currency,$amount);
	 
	
	#Here is the price we are charging...
	
	$paymentNVList{'mo.price'} = $price;
	
	if ($payload ne "") {
	
		# This field is optional... 
		# If you are selling a file as a payload, set this to
		#  the name of the file you wish to sell
	
		$paymentNVList{'mo.payload'} = $payload;
	}
	
	if ($product ne "") {
	
		# product is a description of what is being sold: from the Order Data
		# This is optional
	
		$paymentNVList{'mo.product-descr'} = $product;
	}
	
	if ($orderValidTill ne "") {
	
		# this is a date after which your fulfillment center will 
		# reject the  notification/ fulfillment because the order is too stale.
		#
		# The format of this string is entirely up to you.  If you use
		#  This, you will need to add code to enforce the date in your
		#  fulfillment code.
		
		# This is optional
	
		$paymentNVList{'mo.order-valid-till'} = $orderValidTill;
	}
	
	
	if ($customerId ne "") {
	
		# Some identifier you assigned to identify this customer.
		#  The format is up to you.  This will be passed in the POP.
		
		# This is optional
	
		$paymentNVList{'mo.customer-id'} = $customerId;
	}

	return %paymentNVList;
}

##################################
# ReadPaymentInfo
#     ReadPaymentInfo($orderId);
#
# Inputs:
#     $orderId  -- order identifier... key to recovering payment info
#
#  Returns ($timestamp, $price, $orderValidTill, $payload,
#            $customerId, $product);
#          Notifies, logs then exits if there is a problem
#  
#  This reads order related payment info returns it so that
#  You can display it somehow...
#
# Our sample implementation gets the payment data from a flat file log
# It uses an orderId for the key.  This orderId is passed to the CGI as a query
#  parameter 
#
# Your implementation will probably store the payment data some other way.
###########################

sub ReadPaymentInfo
{
	my ($orderId) = @_;

	my($templatePrms, $errmsg, $errcode);
	my($found, $oid, $timestamp, $price, $orderValidTill, $payload,
        $customerId, $product);

	#Get Order Info from the order log
	#  This uses the test version of an order log
	#  The test version will not scale to large stores.
	#  You will need to replace this function with one that makes 
	#  sense for *your* storefront
	
	($found, $oid, $timestamp, $price, $orderValidTill, $payload,
	        $customerId, $product) = &GetOrderInfo ($orderId);

	if ($found <= 0) {
		# this oid wasn't in the log... you can't continue...
		# Either the input form was tampered with, or you have a fundamental
		# malfunction
	
		if ($found < 0) {  #DB failure....
			$errmsg = &MCKGetErrorMessage($E_No_OrderLog);
			$templatePrms{'MESSAGE'} = $errmsg;
		
			&CCError("$0: $errmsg\n"); 
			$errcode = &PrintTemplate($problemTemplate, %templatePrms);
			if ($errcode != $E_NoErr) {
				&TemplateFailure("$errmsg");
			}
	        exit 1;
		}
		else { # -1 indicates order not found
			$errmsg = &MCKGetErrorMessage($E_Order_NotFound) . " ($orderId)";
			$templatePrms{'MESSAGE'} = $errmsg;
		
			&CCError("$0: $errmsg\n");  #This is not good... 
			$errcode = &PrintTemplate($problemTemplate, %templatePrms);
			if ($errcode != $E_NoErr) {
				&TemplateFailure("$errmsg");
			}
	        exit 1;
		}
	}

	return ($timestamp, $price, $orderValidTill, 
		$payload, $customerId, $product);

}

#  Routines that validate price syntax and manipulate prices
#
#             ******
#
#  Most merchants will never need to edit this code.
# It is in this section because some merchants may handle multiple
# currencies -- and *THEY* will have to test for the acceptable ones here.
#
# Also:  this expects money to be in the format:
#    ccc number
# which is to say, a 3 letter ISO currency code, a blank 
# something recognizable as a valid number.
#
# set the $DECIMAL_POINT variable here to be correct for your locale
#
#  We expect numbers of the form:
#     digits DECIMAL_POINT digits
#     digits
#             DECIMAL_POINT digits
#
#  We will strip anything looking like a thousands separator out of the number.
#
# If you want to control rounding or get fussier, this is the place to do it.
#

#############################
# ValidatePrice
#
# Test the price string for syntactic validity:
# Check for acceptable currency
#  (The currency choices are what you might want to customize...)
#
#############################

$DECIMAL_POINT = '.';

sub ValidatePrice{

	my ($price) = @_;

	my($currency, $amount, $errcode);
	my($thousep) = "[ ,.]";   #chars potentially used as thousands separators
	my(@sepchars);   
	my($ix);

	$errcode = $E_NoErr;

	$price = &Trim($price);

	#if this is null, complain
	if ($price eq "") {
		$errcode = $E_No_Amount;  # no price
		goto BAIL;
	}

	#are there undesirable characters in here?
	if ($price =~ /[^0-9 ,.a-zA-Z ]/) {
		#illegal characters
		$errcode = $E_Bad_Amount;
		goto BAIL;
	}
		
	
	$price =~ tr/A-Z/a-z/;  # lower case it
	$currency = "";

    #  We use $DECIMAL_POINT as a decimal point
    #  We expect numbers of the form:
    #     digits $DECIMAL_POINT digits
    #     digits
    #             $DECIMAL_POINT digits
    #
    #  Potential thousand separators are stripped

	#remove decimal point from thousand separator list
	# This is VERY messy since '.' is special in regular expressions...

	$ix = index($thousep, $DECIMAL_POINT);
	if ($ix >= 0) {
		substr($thousep, $ix, 1) = "";  #delete it from the list
	}
			
	if ($price =~ /^[a-z]/) { #indicates that there should be a currency here
		$price =~ /^([a-z]+)[ \t]*(\S.*)$/;
		$currency = $1;
		$amount = $2;

		if (length($currency) != 3) {
			$errcode = $E_Bad_Currency;
			goto BAIL;
		}
		

		if ($amount  =~ /[^0-9., ]/) {  # with decimal point and separator
			$errcode = $E_Bad_Amount;
			goto BAIL;
		}
	}
	else {
		#No characters at beginning, so no currency
		#Given no currency, is this a number?

		# This allows things like:
		#   100.123
		#   100
		#   .01 

		$amount = $price;

		if ($amount  =~ /[^0-9., ]/) {
        	#illegal characters
        	$errcode = $E_Bad_Amount;
			goto BAIL;
		}
    }

	# strip price and amount of thousands separators

	$amount =~ s/$thousep//g;

	#Finally, is there at most ONE decimal point in here...
    $teststr = "\\$DECIMAL_POINT.*\\$DECIMAL_POINT";

	if ( $amount =~ /$teststr/ ) {
		$errcode = $E_Bad_Amount;
		goto BAIL;
	}
	
	# if there is no currency, use the default.
	if ($currency eq "") {
		# default currency is set with the configuration data
		# Change it there if 'usd' is not acceptable.
    	# use the right ISO 4217 code for your currency

		$currency = $defaultCurrency;
	}

	# This looks like a valid currency code... is the  currency acceptable?
	# At this point, check to see if this currency is acceptable to you...
	# If you accept the default and some others, add them to the test like this:
	#  $currency !~ /$defaultCurrency|frf|cad/ 
	
	if ($currency !~ /$defaultCurrency/ ) 
	{
		# YOICKS! this isn't a currency we can use!
		# bad currency
		$errcode = $E_Reject_Currency;
		goto BAIL;
	}
	
	$price = "$currency $amount";  # canonicalize

BAIL:
	return ($errcode, $price, $currency, $amount);
}


########################################
#  Dump a payload for fulfillment
#  This interprets "payload" as a file name relative to a payload
#  directory
#
#  If this is too restrictive, you can use the field in other
#  ways to suit your needs....
#  
#######################################

sub SendPayload 
{
	my($payloadName) = @_;
	my($mime) = "";
	my($mimeType) = "";

    # Find the extension of payloadName.
    # The extension is everything after the LAST period.
    ($extension) = $payloadName =~ /\.([^.]+?)$/;

    # An extension is required
    unless ($extension) {
        &CCError("$0: ". &MCKGetErrorMessage($E_Bad_Payload) 
			. " ($payloadName)\n");
		return $E_Bad_Payload;
    }

    # Guess the mime type based on the extension. 
    # If it can't find the mime.typ file, just fall through
    open(MIME,"<$Config{\"MIME_TYPE_FILE\"}");
  MIME:
    while (<MIME>) {
        next if /^\#/;     # skip comment lines
        next if /^\s*$/;  # skip blank lines
        ($mimeType,$ext) = /^(.+?)\s+(.+)/;
        my(@ext) = split(' ',$ext);
        foreach (@ext) {
			
            if (/$extension/i)
			{
				$mime = $mimeType;
				last MIME;
			}
        }
    }

    # If mime isn't set, then either:
    #    a) the $mimefile could not be opened, or
    #    b) the extension wasn't listed in the $mimefile
    # Either way, default to text/html
	
    if ($mime eq "")  {
		$mime = 'text/html';
	}

	&CCDebug2(5, "Extension is $extension; Mime type is $mime\n");

	$payloadFile = $Config{"PAYLOAD_DIR"} . $payloadName;
    unless (open(PAYLOAD,"<$payloadFile")) { 
        &CCError("$0: ". &MCKGetErrorMessage($E_No_Payload) 
			. " ($payloadName)\n");
		return $E_No_Payload;
    }
    binmode PAYLOAD;
    binmode STDOUT;

    # find the length of the payload for the progress bar
    $length = -s PAYLOAD;

    print "Content-Type: $mime\r\n";
    print "Content-Length: $length\r\n";
    print "\r\n";
    while (<PAYLOAD>) {
        print;
    }

	return $E_NoErr;  #OK
}

1;
